def gql_locks(fragment):
    return(f'''
query($where: LockWhere!, $first: PageSize!, $skip: Int!) {{
  data: locks(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
''')

GQL_LOCKS_COUNT = f'''
query($where: LockWhere!) {{
  data: countLocks(where: $where)
}}
'''


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from kili import mutations, queries, subscriptions
from kili.authentication import KiliAuth


class Kili(
        mutations.asset.MutationsAsset,
        mutations.label.MutationsLabel,
        mutations.notification.MutationsNotification,
        mutations.organization.MutationsOrganization,
        mutations.project.MutationsProject,
        mutations.user.MutationsUser,
        queries.asset.QueriesAsset,
        queries.label.QueriesLabel,
        queries.lock.QueriesLock,
        queries.organization.QueriesOrganization,
        queries.notification.QueriesNotification,
        queries.project.QueriesProject,
        queries.project_user.QueriesProjectUser,
        queries.user.QueriesUser,
        subscriptions.label.SubscriptionsLabel):

    def __init__(self, api_key=os.getenv('KILI_API_KEY'),
                 api_endpoint='https://cloud.kili-technology.com/api/label/v2/graphql',
                 verify=True):
        """Kili Client
        
        Parameters
        ----------
        api_key: 
            User API key generated from https://cloud.kili-technology.com/label/my-account/api-key

        api_endpoint : str
            Recipient of the HTTP operation
        
        verify : bool
            Verify certificate. Set to False on local deployment without SSL.
        
        Returns
        -------
        Kili object
        Object container your API session.
        Then, list:
        - your assets with: kili.assets()
        - your labels with: kili.labels()
        - your projects with: kili.projects()
        """
        self.auth = KiliAuth(api_key=api_key, api_endpoint=api_endpoint, verify=verify)
        super().__init__(self.auth)



from .fragments import PROJECT_FRAGMENT, PROJECT_FRAGMENT_ID, ROLE_FRAGMENT


GQL_APPEND_TO_ROLES = f'''
mutation($projectID: ID!, $userEmail: String!, $role: ProjectRole!) {{
  data: appendToRoles(
    projectID: $projectID
    userEmail: $userEmail
    role: $role) {{
      {PROJECT_FRAGMENT}
  }}
}}
'''

GQL_UPDATE_PROPERTIES_IN_PROJECT = f'''
mutation(
    $consensusMark: Float
    $consensusTotCoverage: Int
    $description: String
    $honeypotMark: Float
    $instructions: String
    $interfaceCategory: InterfaceCategory
    $inputType: InputType
    $jsonInterface: String
    $minConsensusSize: Int
    $numberOfAssets: Int
    $numberOfAssetsWithSkippedLabels: Int
    $numberOfRemainingAssets: Int
    $numberOfReviewedAssets: Int
    $projectID: ID!
    $title: String
) {{
  data: updatePropertiesInProject(
    where: {{
      id: $projectID
    }},
    data: {{
      consensusMark: $consensusMark
      consensusTotCoverage: $consensusTotCoverage
      description: $description
      honeypotMark: $honeypotMark
      instructions: $instructions
      interfaceCategory: $interfaceCategory
      inputType: $inputType
      jsonInterface: $jsonInterface
      minConsensusSize: $minConsensusSize
      numberOfAssets: $numberOfAssets
      numberOfAssetsWithSkippedLabels: $numberOfAssetsWithSkippedLabels
      numberOfRemainingAssets: $numberOfRemainingAssets
      numberOfReviewedAssets: $numberOfReviewedAssets
      title: $title
    }}
  ) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
'''


GQL_CREATE_EMPTY_PROJECT = f'''
mutation($userID: ID!) {{
  data: createEmptyProject(userID: $userID) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
'''

GQL_CREATE_PROJECT = f'''
mutation(
    $description: String!
    $inputType: InputType!
    $jsonInterface: String!
    $projectType: ProjectType
    $title: String!
    $userID: ID!
  ) {{
  data: createProject(
    description: $description
    inputType: $inputType
    jsonInterface: $jsonInterface
    projectType: $projectType
    title: $title
    userID: $userID) {{
      {PROJECT_FRAGMENT_ID}
  }}
}}
'''

GQL_MAKE_PROJECT_PUBLIC = f'''
mutation(
    $projectID: ID!
  ) {{
  data: makeProjectPublic(
    where: {{
      id: $projectID
  }}) {{
      publicToken
  }}
}}
'''

GQL_UPDATE_PROPERTIES_IN_ROLE = f'''
mutation(
    $roleID: ID!
    $projectID: ID!
    $userID: ID!
    $role: ProjectRole!
) {{
  
  data: updatePropertiesInRole(
    where: {{roleID: $roleID}}
    data: {{
      projectID: $projectID
      userID: $userID
      role: $role
    }}) {{
      {ROLE_FRAGMENT}
  }}
}}
'''

GQL_DELETE_FROM_ROLES = f'''
mutation($roleID: ID!) {{
  data: deleteFromRoles(roleID: $roleID) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
'''

GQL_GQL_UPDATE_PROPERTIES_IN_PROJECT_USER = f'''
mutation(
    $projectUserID: ID!
    $totalDuration: Int
    $numberOfLabeledAssets: Int
    $starred: Boolean
    $consensusMark: Float
    $honeypotMark: Float
) {{
  data: updatePropertiesInProjectUser(
    where: {{id: $projectUserID}},
    data: {{
      totalDuration: $totalDuration
      numberOfLabeledAssets: $numberOfLabeledAssets
      starred: $starred
      consensusMark: $consensusMark
      honeypotMark: $honeypotMark
    }}
  ) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
'''

GQL_DELETE_PROJECT = f'''
mutation($projectID: ID!) {{
  data: deleteProject(where: {{
      id: $projectID
    }}
    ) {{
    {PROJECT_FRAGMENT_ID}
  }}
}}
'''



def gql_notifications(fragment):
    return (f'''
query ($where: NotificationWhere!, $first: PageSize!, $skip: Int!) {{
  data: notifications(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
''')


GQL_NOTIFICATIONS_COUNT = f'''
query($where: NotificationWhere!) {{
  data: countNotifications(where: $where)
}}
'''



import functools
import warnings
import base64
import re
import requests
from json import dumps, loads
from types import *


class Compatible():

    def __init__(self, endpoints=['v1']):
        self.endpoints = endpoints
        self.version_extractor = re.compile(r'\/v\d+/')
        self.address_extractor = re.compile(r':400\d+/')

    def client_is_compatible(self, endpoint):
        version_matched = self.version_extractor.search(endpoint)
        address_matched = self.address_extractor.search(endpoint)
        if not version_matched and not address_matched:
            return False
        if address_matched:
            version = 'v1' if address_matched.group() == ':4000/' else 'v2'
        if version_matched:
            version = 'v1' if version_matched.group() == '/v1/' else 'v2'
        return version in self.endpoints

    def __call__(self, resolver, *args, **kwargs):
        @functools.wraps(resolver)
        def checked_resolver(*args, **kwargs):
            try:
                client_endpoint = args[0].auth.client.endpoint
            except:
                raise ValueError(
                    f'Cannot find client endpoint from resolver {resolver.__name__} with arguments {args}')
            if self.client_is_compatible(client_endpoint):
                return resolver(*args, **kwargs)
            else:
                raise EndpointCompatibilityError(
                    resolver.__name__, client_endpoint)
        return checked_resolver


class EndpointCompatibilityError(Exception):
    def __init__(self, resolver, endpoint):
        super().__init__(
            f'Resolver {resolver} is not compatible with the following endpoint : {endpoint}')


class GraphQLError(Exception):
    def __init__(self, mutation, error):
        super().__init__(f'Mutation "{mutation}" failed with error: "{error}"')


def format_result(name, result, Object=None):
    if 'errors' in result:
        raise GraphQLError(name, result['errors'])
    formatted_json = format_json(result['data'][name])
    if Object is None:
        return formatted_json
    if isinstance(formatted_json, list):
        return [Object(element) for element in formatted_json]
    return Object(formatted_json)


def content_escape(content):
    return content.replace('\\', '\\\\').replace('\n', '\\n').replace('"', '\\"')


def get_data_type(path):
    if re.match(r'.*(jpg|jpeg)$', path.lower()):
        return 'image/png'
    if re.match(r'.*pdf$', path.lower()):
        return 'application/pdf'


def encode_base64(path):
    data_type = get_data_type(path)
    with open(path, 'rb') as image_file:
        return f'data:{data_type};base64,' + \
            base64.b64encode(image_file.read()).decode('ascii')


def is_url(path):
    return isinstance(path, str) and re.match(r'^(http://|https://)', path.lower())


def format_json(result):
    if result is None:
        return result
    if isinstance(result, list):
        return [format_json(elem) for elem in result]
    if isinstance(result, dict):
        for key, value in result.items():
            if key in ['jsonInterface', 'jsonMetadata', 'jsonResponse']:
                if (value == '' or value is None) and not (is_url(value) and key == 'jsonInterface'):
                    result[key] = dict()
                else:
                    try:
                        if is_url(value):
                            result[key] = requests.get(value).json()
                        else:
                            result[key] = loads(value)
                    except:
                        raise ValueError(
                            'Json Metadata / json response / json interface should be valid jsons')
            else:
                result[key] = format_json(value)
        return result
    return result


def fragment_builder(fields, type_of_fields):
    fragment = ''
    subfields = [field.split('.', 1) for field in fields if '.' in field]
    if subfields:
        for subquery in set([subfield[0] for subfield in subfields]):
            type_of_fields_subquery = getattr(type_of_fields, subquery)
            try:
                if issubclass(type_of_fields_subquery, object):
                    fields_subquery = [subfield[1]
                                       for subfield in subfields if subfield[0] == subquery]
                    fragment += f' {subquery}{{{fragment_builder(fields_subquery,type_of_fields_subquery)}}}'
            except ValueError:
                print(f'{subquery} must be a valid subquery field')
        fields = [field for field in fields if '.' not in field]
    for field in fields:
        try:
            getattr(type_of_fields, field)
        except ValueError:
            print(f'{field} must be an instance of {type_of_fields}')
        if isinstance(field, str):
            fragment += f' {field}'
        else:
            raise Exception('Please provide the fields to query as strings')
    return fragment


def deprecate(msg, type=DeprecationWarning):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(msg, type, stacklevel=2)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def format_metadata(metadata):
    if metadata is None:
        return metadata
    elif isinstance(metadata, str):
        return metadata
    elif isinstance(metadata, dict) or isinstance(metadata, list):
        return dumps(metadata)
    else:
        raise Exception(
            f"Metadata {metadata} of type {type(metadata)} must either be None, a string a list or a dict.")


def convert_to_list_of_none(array, length):
    if isinstance(array, list):
        if len(array) != length:
            raise Exception(f'array should have length {length}')
        else:
            return array
    else:
        return [None] * length


def is_none_or_empty(object):
    object_is_empty = isinstance(object, list) and len(object) == 0
    return object is None or object_is_empty


def list_is_not_none_else_none(object):
    return [object] if object is not None else None


def infer_id_from_external_id(playground, asset_id: str, external_id: str, project_id: str):
    if asset_id is None and external_id is None:
        raise Exception(
            'Either provide asset_id or external_id and project_id')
    if asset_id is not None:
        return asset_id
    assets = playground.assets(
        external_id_contains=[external_id], project_id=project_id, fields=['id'], disable_tqdm=True)
    if len(assets) == 0:
        raise Exception(
            f'No asset found with external ID "{external_id}"')
    if len(assets) > 1:
        raise Exception(
            f'Several assets found containing external ID "{external_id}": {assets}. Please, use asset ID instead.')
    return assets[0]['id']



from .fragments import AUTH_PAYLOAD_FRAGMENT, USER_FRAGMENT

GQL_SIGN_IN = f'''
mutation($email: String!, $password: String!) {{
  data: signIn(email: $email password: $password) {{
    {AUTH_PAYLOAD_FRAGMENT}
  }}
}}
'''

GQL_CREATE_USER = f'''
mutation(
    $name: String!
    $email: String!
    $password: String!
    $organizationRole: OrganizationRole!
) {{
  data: createUser(name: $name
      email: $email
      password: $password
      organizationRole: $organizationRole) {{
    {USER_FRAGMENT}
  }}
}}
'''

GQL_CREATE_USER_FROM_EMAIL_IF_NOT_EXISTS = f'''
mutation(
    $name: String!
    $email: String!
    $organizationRole: OrganizationRole!
    $projectID: String!
) {{
  data: createUserFromEmailIfNotExists(name: $name
      email: $email
      organizationRole: $organizationRole
      projectID: $projectID) {{
    {USER_FRAGMENT}
  }}
}}
'''

GQL_UPDATE_API_KEY = f'''
mutation(
    $email: String!
    $newApiKey: String!
) {{
  data: updateApiKey(email: $email
      newApiKey: $newApiKey) {{
    {USER_FRAGMENT}
  }}
}}
'''

GQL_UPDATE_PASSWORD = f'''
mutation(
    $email: String!
    $oldPassword: String!
    $newPassword1: String!
    $newPassword2: String!
) {{
  data: updatePassword(email: $email
      oldPassword: $oldPassword
      newPassword1: $newPassword1
      newPassword2: $newPassword2) {{
    {USER_FRAGMENT}
  }}
}}
'''

GQL_RESET_PASSWORD = f'''
mutation($email: String!) {{
  data: resetPassword(email: $email) {{
    {USER_FRAGMENT}
  }}
}}
'''

GQL_UPDATE_PROPERTIES_IN_USER = f'''
mutation(
    $email: String!
    $name: String
    $organizationId: String
    $organizationRole: OrganizationRole
    $activated: Boolean
) {{
  data: updatePropertiesInUser(
    where: {{email: $email}}
    data: {{
      name: $name
      email: $email
      organizationId: $organizationId
      organizationRole: $organizationRole
      activated: $activated
    }}
  ) {{
    {USER_FRAGMENT}
  }}
}}
'''



def gql_labels(fragment):
    return(f'''
query ($where: LabelWhere!, $first: PageSize!, $skip: Int!) {{
  data: labels(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
''')


GQL_LABELS_COUNT = f'''
query($where: LabelWhere!) {{
  data: countLabels(where: $where)
}}
'''



from .fragments import LABEL_FRAGMENT, LABEL_FRAGMENT_ID

GQL_CREATE_PREDICTIONS = f'''
mutation(
    $projectID: ID!
    $externalIDArray: [ID!]!
    $modelNameArray: [String!]!
    $jsonResponseArray: [String!]!
) {{
  data: createPredictions(
    projectID: $projectID
    externalIDArray: $externalIDArray
    modelNameArray: $modelNameArray
    jsonResponseArray: $jsonResponseArray) {{
      {LABEL_FRAGMENT_ID}
  }}
}}
'''

GQL_APPEND_TO_LABELS = f'''
mutation(
    $authorID: ID!
    $jsonResponse: String!
    $labelAssetID: ID!
    $labelType: LabelType!
    $secondsToLabel: Float
    $skipped: Boolean!
) {{
  data: appendToLabels(
    authorID: $authorID
    jsonResponse: $jsonResponse
    labelAssetID: $labelAssetID
    labelType: $labelType
    secondsToLabel: $secondsToLabel
    skipped: $skipped) {{
      {LABEL_FRAGMENT_ID}
  }}
}}
'''


GQL_UPDATE_PROPERTIES_IN_LABEL = f'''
mutation(
    $labelID: ID!
    $secondsToLabel: Float
    $modelName: String
    $jsonResponse: String
) {{
  data: updatePropertiesInLabel(
    where: {{id: $labelID}}
    data: {{
      secondsToLabel: $secondsToLabel
      modelName: $modelName
      jsonResponse: $jsonResponse
    }}
  ) {{
    {LABEL_FRAGMENT_ID}
  }}
}}
'''

GQL_CREATE_HONEYPOT = f'''
mutation(
    $assetID: ID!
    $jsonResponse: String!
) {{
  data: createHoneypot(
    assetID: $assetID
    jsonResponse: $jsonResponse) {{
      {LABEL_FRAGMENT}
  }}
}}
'''
